---------------------------------------------------------------------
-- TITLE: cpu
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 1.3.2012
-- FILENAME: MAIS_soc.vhd
-- PROJECT: MAIS CPU with SPB lite Bus
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--
-- DESCRIPTION:

---------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use ieee.numeric_std.all;


use work.SPB.all;
use work.SPBdevice.all;


entity MAIS_soc is
  generic(debug_file    : string                        := "no";
          sim           : string                        := "no";
          ISR_addr      : std_logic_vector(31 downto 0) := X"0000004C";
          SYS_addr      : std_logic_vector(31 downto 0) := X"00000100";            
          board_clk_frq : integer                       := 27E6;
      --    multiplier    : string                      := "none");
          multiplier    : string  := "generic");

  port (board_clk : in  std_logic;
        reset     : in  std_logic := '1';
        HW_INT    : in  std_logic;
        
        --UART
        rx        : in  std_logic := '1';
        tx        : out std_logic);
end MAIS_soc;

architecture Behavioral of MAIS_soc is
  constant IRQ_number       : positive := 2; --HW_INT and Uart_IRQ
  constant number_of_slaves : natural  := 3; --Memory Uart Busdummy
  constant sel_size         : natural  :=16;
  
  type     slaveback is array (0 to number_of_slaves-1) of SPB_out;

  signal slave         : slaveback;
  signal BBUS_in       : SPB_in;
  signal BBUS_out      : SPB_out;
  signal IRQ           : std_logic_vector (IRQ_number-1 downto 0);
  signal BSEL          : std_logic_vector (sel_size-1 downto 0);

  signal imem_data     : std_logic_vector(31 downto 0);
  signal imem_valid    : std_logic;
  signal imem_address  : std_logic_vector(31 downto 0);
  
  signal IRQ_UART      : std_logic;

  signal reset_counter : unsigned (3 downto 0) := (others => '0');
  signal Breset        :std_logic;

--Memory  Map
--0x00000000    RAM
--0x20000000    UART

begin
  IRQ <= HW_INT & IRQ_UART;
  
---BUS multiplexer  
    BBUS_out <= slave(0) when BSEL(0)='1' else  --DRAM
    slave(1)             when BSEL(2)='1' else  --UART
    slave(2);   --Bus Dummy   

  
  CPU : MAIS_CPU generic map (
    IRQ_number     => IRQ_number,
    ISR_addr       => ISR_addr,
    SYS_addr       => SYS_addr,
    multiplier     => multiplier
    )

    port map(
      bclk         => board_clk,
      breset       => breset,
      IRQ          => IRQ,
      imem_data    => imem_data,
      imem_valid   => imem_valid,
      imem_address => imem_address,
      ---SPB BUS
      BBUS_in      => BBUS_in,
      BBUS_out     => BBUS_out
      );


  --SLAVE 0 -- DRAM
  Inst_memory : SPB_memory port map(
    BCLK           => board_clk,
    BBUS_in        => BBUS_in,
    BBUS_out       => slave(0),
    imem_address   => imem_address,
    imem_data      => imem_data,
    imem_valid     => imem_valid
    );


  --SLAVE 1 -- UART
  HB_UART : SPB_UART  generic map(
    sim            => sim,
    BSEL           => 2,
	baudrate       => 115200,
    clk_freq       => board_clk_frq)

  port map(
    BCLK          => board_clk,
    BRESET        => BRESET,
    BBUS_in       => BBUS_in,
    BBUS_out      => slave(1),

    uart_irq      => IRQ_UART,
    rx            => rx,
    tx            => tx);


  --HSlave2 Dummy            
  HB_dummy : bus_dummy  generic map(
    SEL         => 15)
 
  port map(
    BCLK          => board_clk,
    BRESET        => BRESET,
    BBUS_in       => BBUS_in,
    BBUS_out      => slave(2));

  

  process(board_clk)
  begin
    if rising_edge (board_clk) then
      if reset_counter /= X"F" then
        reset_counter <= reset_counter+1;
      end if;
      if reset = '1' then
        reset_counter <= (others => '0');
      end if;
    end if;
  end process;

  process(board_clk)
  begin
    if rising_edge (board_clk) then
      if reset_counter = X"F" then
        breset <= '0';
      else
        breset <= '1';
      end if;
    end if;
  end process;

  process(board_clk)
  begin
    if rising_edge (board_clk) then
      if BBUS_in.BREADY = '1' then
        BSEL <= BBUS_in.BSEL;
      end if;
      if reset = '1' then
        BSEL <= (others => '0');
      end if;
    end if;
  end process;

 
end Behavioral;

