---------------------------------------------------------------------
-- TITLE: Coprocessor 0 (system processor)
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 10.7.2011
-- FILENAME: cp0.vhd
-- PROJECT: MAIS CPU core
-- COPYRIGHT: see Mais Processor Manual
-- --
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--
-- DESCRIPTION :
-- Implements  coprocessor
---            Interrupt logic
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mlite_pack.all;

entity cp0 is
  generic (IRQ_number : integer);
  port (
    cpu_clk            : in std_logic;
    reset              : in std_logic;
    PC_reg             : in std_logic_vector(31 downto 0);
    PC_plus4           : in std_logic_vector(31 downto 0);
    syscall            : in std_logic;
    restore            : in std_logic;
    IRQ                : in std_logic_vector(IRQ_number-1 downto 0);
    break1             : in std_logic;
    branch             : in branch_rec;
    imem_valid         : in std_logic;
    fetch              : in     fetch_rec;
    exception          : buffer std_logic := '0';
    exception_pipeline : buffer unsigned (2 downto 0);
    CP0_write          : in     write_reg_rec;  --execute   

    jump_ISR           : buffer std_logic;
    debug_pc_hold      : out    std_logic; 
    debug_fetch_hold   : out    std_logic;
 


    CP0_out            : out CP0_rec;
    CPint              : out CPint_rec
    );
end cp0;


  
architecture behavioral of cp0 is

--CPO 
--register  9   Count --counter
--register 13   Cause --interrupt exception
--register 14   EPC   --Exception Program Counter
--register 15   SYSPC --SYSTEM Program Counter  for OS

--register 23   Debug --can used by JTAG
--register 24   DEPC
  signal reg9  : std_logic_vector (31 downto 0) := X"0000_0000";
  signal reg11 : std_logic_vector (31 downto 0) := X"0000_0000";
  signal reg12 : std_logic_vector (31 downto 0) := X"0000_0000";
  signal reg13 : std_logic_vector (31 downto 0) := X"0000_0000";
  signal reg14 : std_logic_vector (31 downto 0);  --:=X"0000_0000";  
  signal reg15 : std_logic_vector (31 downto 0);
  signal reg23 : std_logic_vector (31 downto 0) := X"0000_0000";



  alias Count   : std_logic_vector (31 downto 0) is reg9;
  alias Compare : std_logic_vector (31 downto 0) is reg11;
  alias status  : std_logic_vector (31 downto 0) is reg12;
  alias cause   : std_logic_vector (31 downto 0) is reg13;
  alias EPC     : std_logic_vector (31 downto 2) is reg14(31 downto 2);
  alias SYSPC   : std_logic_vector (31 downto 2) is reg15(31 downto 2);

 -- signal exception_d1 : std_logic := '0';
  signal ISR_runs     : std_logic;
  signal restore_d1   : std_logic;
  signal pc_hold      : std_logic;

begin

  CPint.EPC <= EPC;



  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      restore_d1<=restore;
      if reset = '1' then
        restore_d1<='0';     
      end if;
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if fetch.IR_new = '1' then
        if exception_pipeline /= "000" then
          exception_pipeline <= exception_pipeline-1;
        end if;
      end if;
      if jump_isr = '1' then
        exception_pipeline <= "100";
      end if;
      if reset = '1' then
        exception_pipeline <= (others => '0');
      end if;
    end if;
  end process;

--  debug_hold <= reg23(30) and  not ISR_runs;  

-- inside ISR?
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if jump_isr = '1' then
        ISR_runs <= '1';
        pc_hold<='0';        
      end if;
           
      --(return from interrupt service routine)
      if restore = '1' then
        ISR_runs <= '0';
      end if;

      if imem_valid='1' and pc_hold='1' then
        debug_fetch_hold<='1';
      else
        debug_fetch_hold<='0';      
      end if;
      if (restore='1' and reg23(30)='1') then     
        pc_hold<='1';   
      end if;
      if (restore_d1='1' and reg23(29)='1') then  
        pc_hold<='1';              
      end if;
      
      if reset = '1' then
        pc_hold<='0';
        debug_fetch_hold<='0'; 
         
        ISR_runs <= '0';    
      end if;     
    end if;
  end process;

  debug_pc_hold<=pc_hold;

--- hier nochmal Nachschauen ganz wichtig
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if status(0) = '1' then
        if (IRQ /= std_logic_vector(to_unsigned(00,IRQ'left +1))
                or reg23(1) = '1'or reg23(29)='1') and ISR_runs = '0' then
          exception <= '1';
          if exception = '0' then
            jump_isr <= '1';
          else
            jump_isr <= '0';
          end if;
        end if;
      end if;
      if fetch.ir_new = '1' and exception = '1' then
        exception <= '0';
        jump_isr  <= '0';
        
      end if;

      if reset = '1' then
        exception <= '0';
        jump_ISR  <= '0';
      end if;
    end if;
  end process;





  cycle_count : process(cpu_clk)        --reg9
  begin
    if rising_edge(cpu_clk) then
      if reset = '1' then
        Count <= (others => '0');
      else
        Count <= std_logic_vector(unsigned(Count)+1);
      end if;
    end if;
  end process;


-- bit 0  Interrupt maske
  process(cpu_clk)  -- reg 12  status     reg12(0)-- INT enable
  begin
    if rising_edge(cpu_clk) then
      if CP0_write.wr = '1' then
        if CP0_write.index = "01100" then
          reg12 <= CP0_write.data;
        end if;
      end if;
      if reset = '1' then
        status <= (others => '0');
      end if;
    end if;
  end process;


-- cause reg 13
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if CP0_write.wr = '1' then
        if CP0_write.index = "01101" then
          reg13 <= CP0_write.data;
        end if;
      end if;

      reg13(31) <= reg23(30);

      if exception = '1' then
        reg13(4) <= '1';
      end if;
    end if;
  end process;



-- EPC reg 14
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if CP0_write.wr = '1' then
        if CP0_write.index = "01110" then
          reg14 <= CP0_write.data;
        end if;
      end if;
      if jump_isr = '1' then
		  reg14 <= PC_plus4;
		  if break1='1' or reg23(30)='1' then
		    reg14 <=pc_reg;
		  end if;
        if branch.branch_execute = '1' then
          reg14 <= branch.branch_address & "00";
        end if;
      end if;
      if exception_pipeline /= "000" and branch.branch_execute='1' and syscall='0' then
        reg14 <= branch.branch_address & "00";
      end if;
    end if;
  end process;



-- SYSPC reg 15
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if syscall = '1' then
        reg15 <= fetch.pc_d1 &"00";
      end if;
    end if;
  end process;

--debug  --reg23
--
--reg23(31) allow Debug break exeption
--reg23(30) Debug mode  stop PC outside ISR
--reg23(29) allow Debug single step



--reg23(1)  Debug breakpoint interrupt
--reg23(0)  Debug single step
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if CP0_write.wr = '1' then
        if CP0_write.index = "10111" then  --reg23(31) enable INT break1
          reg23(31 downto 0) <= CP0_write.data(31 downto 0);  --reg23(30) Debug mode jtag spec stop PC
        end if;
      end if;
      
      if break1 = '1' and reg23(31) = '1' then
        reg23(1) <= '1';                --reg23(1) Debug breakpoint exeption
      end if;
      
      if  ISR_Runs='0' then
        reg23(0)<=reg23(29);
      end if;      
 
      if reset = '1' then
        reg23 <= X"00000000";
      end if;
    end if;
  end process;



  reg_out : process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      --    case CP0_index is
      case fetch.IR(15 downto 11) is
        when "01001" => CP0_out.reg <= Count;
        when "01101" => CP0_out.reg <= reg13;  --Cause         
        when "01110" => CP0_out.reg <= reg14;  --EPC
        when "01111" => CP0_out.reg <= reg15;  --SYSPC
        when "10111" => CP0_out.reg <= reg23;
        when others  => null;
      end case;
    end if;
  end process;

  
  
end behavioral;





