---------------------------------------------------------------------
-- TITLE: cpu
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 27.6.2011
-- FILENAME: MAIS_cpu.vhd
-- PROJECT: MAIS CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
-- 
-- DESCRIPTION:
--    connect the pipeline stage together
--    connect the I-RAM and the D-RAM
---------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use ieee.numeric_std.all;
use work.mlite_pack.all;
use work.SPB.all;

entity MAIS_CPU is
  generic (IRQ_number : integer;
           ISR_addr   : std_logic_vector(31 downto 0);
           SYS_addr   : std_logic_vector(31 downto 0);
           multiplier : string := "generic");

  port (bclk          : in std_logic;
        breset        : in std_logic;
        IRQ           : in std_logic_vector(IRQ_number-1 downto 0);

        ---Instruction RAM
        imem_data    : in  std_logic_vector(31 downto 0);
        imem_valid   : in  std_logic;
        imem_address : out std_logic_vector(31 downto 0);

        ---SPBlite BUS
        BBUS_in  : out SPB_in;
        BBUS_out : in  SPB_out
        );
end MAIS_CPU;


architecture Behavioral of MAIS_CPU is


  constant baudrate         : integer   := 115200;
  signal   reset_a, reset_b : std_logic := '0';  --! reset catcher

  component inst_pc is
    generic (
      ISR_addr : std_logic_vector(31 downto 0);
      SYS_addr : std_logic_vector(31 downto 0));    
    port(cpu_clk            : in  std_logic;   --! cpu clock
         reset              : in  std_logic;   --! reset PC   
         jump_ISR           : in  std_logic;
         syscall            : in  std_logic;
         op_break           : in  std_logic;   --! PC working 
         stall              : in  std_logic;   --! PC working 
         stall_mult         : in  std_logic;
         debug_pc_hold      : in  std_logic;
         breakC             : in  std_logic;   --! PC working 
         exception_pipeline : in  unsigned (2 downto 0);
         BREADY             : in  std_logic;   --! READY SPB-lite
         BACTIVE            : in  std_logic;
         imem_valid         : in  std_logic;
         branch             : in  branch_rec;  --! control the program counter
         PC_reg             : out std_logic_vector(31 downto 0);
         PC_plus4           : out std_logic_vector(31 downto 0);
         PC_plus4_d1        : out std_logic_vector(31 downto 0)
         );
  end component;


  signal PC_reg      : std_logic_vector(31 downto 0);
  signal PC_plus4    : std_logic_vector(31 downto 0);
  signal PC_plus4_d1 : std_logic_vector(31 downto 0);


  component inst_fetch is
    port(
      cpu_clk            : in std_logic;
      reset              : in std_logic;
      syscall            : in std_logic;
      break7             : in std_logic;
      exception          : in std_logic;
      exception_pipeline : in unsigned (2 downto 0);
      bus_wait           : in std_logic;
      BACTIVE            : in std_logic;
      BREADY             : in std_logic;

      pc_reg             : in std_logic_vector(31 downto 0);
      pc_plus4_d1        : in std_logic_vector(31 downto 0);

      imem_data          : in std_logic_vector(31 downto 0);
      imem_valid         : in std_logic;

      mult_ready         : in     std_logic;
      branch             : in     branch_rec;
      stall              : in     std_logic;
      stall_mult         : buffer std_logic;

      op_break           : in std_logic;
      debug_fetch_hold   : in std_logic;  
      fetch              : out fetch_rec
      );
  end component;

  signal stall      : std_logic;
  signal stall_mult : std_logic;
  signal fetch      : fetch_rec;


  component inst_decode is
    port(
      cpu_clk            : in     std_logic;
      reset              : in     std_logic;
      exception_pipeline : in     unsigned (2 downto 0);
      branch_execute     : in     std_logic;
      fetch              : in     fetch_rec;
      load_wr            : in     std_logic;  ---?
      BACTIVE            : in     std_logic;
      BREADY             : in     std_logic;
      execute_reg        : in     write_reg_rec;
      op_break           : buffer std_logic;
      break1             : out    std_logic;
      break7             : out    std_logic;
      stall              : in     std_logic;
      stall_mult         : in     std_logic;
      decode             : out    decode_rec
      );
  end component;

  signal decode         : decode_rec;
  signal decode_disable : std_logic;
  signal op_break       : std_logic;
  signal break1         : std_logic; 
  signal break7         : std_logic;


  component inst_execute is
  generic(multiplier : string := "generic");
  port (

    cpu_clk     : in  std_logic;
    mult_ready  : out std_logic := '0';
    reset       : in  std_logic;
    stall       : in  std_logic;
    BREADY      : in  std_logic;
    decode      : in  decode_rec;
    loadto_reg  : in  write_reg_rec;
    --CP0

    CP0_out     : in  CP0_rec;
    CPint       : in  CPint_rec;
    CP0_write   : out write_reg_rec;

    --fetch    
    branch      : out branch_rec;    --! control the program counter
    execute_reg : out write_reg_rec;  
    --load_access

    IR_load     : out std_logic_vector (2 downto 0);
    load_index  : out std_logic_vector (4 downto 0);

    --BBUS
    bus_wait    : in     std_logic;
    BADDR       : out    std_logic_vector (31 downto 0);
    BSEL        : out    std_logic_vector (sel_size-1 downto 0);
    BRD         : out    std_logic;
    BWR         : out    std_logic;
    BACTIVE     : out    std_logic;
    BMASK       : out    std_logic_vector (3 downto 0);
    BWDATA      : out    std_logic_vector (31 downto 0)
    ); 
  end component;


  signal execute_reg : write_reg_rec;
  signal CP0_write   : write_reg_rec;
  signal IR_load     : std_logic_vector (2 downto 0);
  signal load_index  : std_logic_vector(4 downto 0);
  signal branch      : branch_rec;
  signal mult_ready  : std_logic;

  signal BBUS_a : SPB_in;


  component load_access is
    port (
      cpu_clk    : in  std_logic;
      reset      : in  std_logic;
      BADDR      : in  std_logic_vector (31 downto 0);
      BWDATA     : in  std_logic_vector (31 downto 0);
      BRD        : in  std_logic;
      BRDATA     : in  std_logic_vector (31 downto 0);
      BREADY     : in  std_logic;
      --from execute
      load_index : in  std_logic_vector (4 downto 0);
      IR_load    : in  std_logic_vector (2 downto 0);
      --output  
      loadto_reg : out write_reg_rec
      );
  end component;


  signal loadto_reg : write_reg_rec;

  component cp0 is
    generic (IRQ_number : positive);
    port (
      cpu_clk            : in std_logic;
      reset              : in std_logic;
      PC_reg             : in std_logic_vector(31 downto 0);
      PC_plus4           : in std_logic_vector(31 downto 0);
      syscall            : in std_logic;
      restore            : in std_logic;
      IRQ                : in std_logic_vector(IRQ_number-1 downto 0);
      break1             : in std_logic;
      branch             : in branch_rec;
      imem_valid         : in std_logic;

      fetch              : in fetch_rec;

      exception          : buffer std_logic;
      exception_pipeline : buffer unsigned (2 downto 0);

      CP0_write          : in     write_reg_rec;  --execute
      jump_ISR           : buffer std_logic;
      debug_pc_hold      : out    std_logic;
      debug_fetch_hold   : out    std_logic;

      CP0_out            : out    CP0_rec;
      CPint              : out    CPint_rec
      );
  end component;

  signal jump_ISR           : std_logic;
  signal debug_pc_hold      : std_logic;
  signal debug_fetch_hold   : std_logic;
  signal CP0_out            : CP0_rec;
  signal CPint              : CPint_rec;
  signal exception          : std_logic;
  signal exception_pipeline : unsigned (2 downto 0);

  component inst_arbiter is
    port(
      cpu_clk        : in     std_logic;
      reset          : in     std_logic;
      BREADY         : in     std_logic;
      BACTIVE        : in     std_logic;
      decode         : in     decode_rec;
      bus_wait       : buffer std_logic;

      stall : buffer std_logic
      );
  end component;

  signal bus_wait : std_logic;


  
begin



  
  p1_pc : inst_pc generic map(
    ISR_addr => ISR_addr,
    SYS_addr => SYS_addr)
    port map (
      cpu_clk            => BCLK,
      reset              => reset_b,
      jump_ISR           => jump_ISR,
      syscall            => decode.syscall,
      op_break           => op_break,
      stall              => stall,
      stall_mult         => stall_mult,
      debug_pc_hold      => debug_pc_hold,
      breakC             => stall_mult,
      exception_pipeline => exception_pipeline,
      BREADY             => BBUS_OUT.BREADYOUT,
      BACTIVE            => BBUS_a.BACTIVE,
      imem_valid         => imem_valid,
      branch             => branch,
      PC_reg             => PC_reg,
      PC_plus4           => PC_plus4,
      PC_plus4_d1        => PC_plus4_d1
      );

  imem_address <= PC_reg;
 

  p1_fetch : inst_fetch port map(
    cpu_clk            => BCLK,
    reset              => reset_a,
    syscall            => decode.syscall,
    break7             => break7,
    exception          => exception,
    exception_pipeline => exception_pipeline,
    bus_wait           => bus_wait,
    BACTIVE            => BBUS_a.BACTIVE,
    BREADY             => BBUS_OUT.BREADYOUT,
    pc_reg             => PC_reg,
    PC_plus4_d1        => PC_plus4_d1,
    imem_data          => imem_data,
    imem_valid         => imem_valid,
    mult_ready         => mult_ready,
    branch             => branch,
    stall              => stall,
    stall_mult         => stall_mult,
    op_break           => op_break,
    debug_fetch_hold   => debug_fetch_hold,
    fetch              => fetch
    );


  p2_decode : inst_decode port map(
    cpu_clk            => BCLK,
    reset              => reset_b,
    exception_pipeline => exception_pipeline ,
    branch_execute     => branch.branch_execute,

    fetch              => fetch,
    load_wr            => loadto_reg.wr,
    BACTIVE            => BBUS_a.BACTIVE,
    BREADY             => BBUS_OUT.BREADYOUT,
    execute_reg        => execute_reg,
    op_break           => op_break,
    break1             => break1,
    break7             => break7,
    stall              => stall,
    stall_mult         => stall_mult,
    decode             => decode
    );

  
  p3_execute : inst_execute
    generic map(
      multiplier => multiplier)

    port map (
      cpu_clk    => BCLK,
      mult_ready => mult_ready,
      reset      => reset_b,
      stall      => stall,
      BREADY     => BBUS_OUT.BREADYOUT,
      decode     => decode,
      loadto_reg => loadto_reg,

      CP0_out   => CP0_out,
      cpint     => cpint,
      CP0_write => CP0_write,

      branch => branch,

      execute_reg => execute_reg,
      load_index  => load_index,
      IR_load     => IR_load,
      bus_wait    => bus_wait,
      BADDR       => BBUS_a.BADDR,
      BSEL        => BBUS_a.BSEL,
      BRD         => BBUS_a.BRD,
      BWR         => BBUS_a.BWR,
      BACTIVE     => BBUS_A.BACTIVE,
      BMASK       => BBUS_a.BMASK,
      BWDATA      => BBUS_a.BWDATA
      );  


  p5_writeback : load_access port map (
    --BBUS
    cpu_clk => BCLK,
    reset   => reset_b,

    BADDR      => BBUS_a.BADDR,
    BWDATA     => BBUS_a.BWDATA,
    BRD        => BBUS_a.BRD,
    BRDATA     => BBUS_out.BRDATA,
    BREADY     => BBUS_out.BREADYOUT,
 
    --from execute
    load_index => load_index,
    IR_load    => IR_load,
    --output
    loadto_reg => loadto_reg
    );

  
    BBUS_a.BREADY <= BBUS_out.BREADYOUT;



  BBUS_in <= BBUS_a;

  bus_arbiter : inst_arbiter
    port map(
      cpu_clk        => BCLK,
      reset          => reset_a,
      BREADY         => BBUS_out.BREADYOUT,
      BACTIVE        => BBUS_a.BACTIVE,
      decode         => decode,
      bus_wait       => bus_wait,
      stall          => stall
      );




  a0_cp0 : cp0
    generic map(
      IRQ_number         => IRQ_number) 
    port map (
      cpu_clk            => BCLK,
      reset              => reset_a,
      PC_reg             => PC_reg,
      PC_plus4           => PC_plus4,
      syscall            => decode.syscall,
      restore            => decode.restore,
      IRQ                => IRQ,
      break1             => break1,
      branch             => branch,
      imem_valid         => imem_valid,
      fetch              => fetch,

      exception          => exception,
      exception_pipeline => exception_pipeline,
      CP0_write          => CP0_write,
      jump_ISR           => jump_ISR,
      debug_pc_hold      => debug_pc_hold,
      debug_fetch_hold   => debug_fetch_hold,
      CP0_out            => CP0_out,
      CPint              => CPint
      );

  ------------
  process(BCLK)
  begin
    if rising_edge(BCLK) then
      reset_a <= breset;
      reset_b <= reset_a;
    end if;
  end process;





end Behavioral;

