---------------------------------------------------------------------
-- TITLE: Instruction decode
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 2/8/2011
-- FILENAME: inst_decode.vhd
-- PROJECT: Mais CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--  
-- DESCRIPTION:
--    implements pipeline stage decode
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mlite_pack.all;
use work.control_pack.all;

entity inst_decode is

  port (
    cpu_clk            : in     std_logic;
    reset              : in     std_logic;
    exception_pipeline : in     unsigned (2 downto 0);
    branch_execute     : in     std_logic;
    fetch              : in     fetch_rec;
    load_wr            : in     std_logic;
    BACTIVE            : in     std_logic;
    BREADY             : in     std_logic;
    execute_reg        : in     write_reg_rec;
    op_break           : buffer std_logic;
    break1             : out    std_logic;
    break7             : out    std_logic;
    stall              : in     std_logic;
    stall_mult         : in     std_logic;
    decode             : out    decode_rec);   

end inst_decode;



architecture Behavioral of inst_decode is


  
  signal control_out   : control_rec;
  signal stall_d1      : std_logic;


  component reg_bank
    port(
      cpu_clk     : in  std_logic;
      stall       : in  std_logic;
      BREADY      : in  std_logic;
      rs_index    : in  std_logic_vector(4 downto 0);
      rt_index    : in  std_logic_vector(4 downto 0);
      execute_reg : in  write_reg_rec;
      regrs_out   : out std_logic_vector(31 downto 0);
      regrt_out   : out std_logic_vector(31 downto 0)
      );
  end component;


begin

--if instruction break or syscall in IR?
-- then stop execution 
  op_break <= '1' when fetch.IR(31 downto 26) = "000000" and fetch.IR(5 downto 1) = "00110"
              else '0';
  decode.syscall <= '1' when op_break = '1' and fetch.IR(0) = '0' else '0';

--break1 is reserved for debugger
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if op_break = '1' and fetch.IR(19 downto 16) = "0001" then
        break1 <= '1';
      else
        break1 <= '0';
      end if;
    end if;
  end process;
  
--break7 is break multiplier
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if op_break = '1' and fetch.IR(19 downto 16) = "0111" then
        break7 <= '1';
      else
        break7 <= '0';
      end if;
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if control_out.restore = '1' then
        decode.restore <= '1';
      else
        decode.restore <= '0';
      end if;
      if reset = '1' then
        decode.restore <= '0';
      end if;
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      stall_d1      <= stall;


   
      if stall = '0' and stall_mult = '0' then
        decode.PC_branch <= std_logic_vector(signed(fetch.pc_plus4_d2)
                                             + signed(fetch.IR(15 downto 0)));        
      end if;
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then

     
      if stall = '0' and not (BACTIVE = '1' and BREADY = '0') then
        control_out  <= control_f(fetch.IR);
        decode.IR_d1 <= fetch.IR;
        decode.imm32 <= std_logic_vector(resize(signed(fetch.IR(15 downto 0)), 32));
        --logical immediate operation
        if fetch.IR(29 downto 28) = "11" then
          decode.imm32(31 downto 16) <= X"0000";  --immediate unsigned 
        end if;
      end if;

      if stall = '0' and (BACTIVE = '0' or BREADY = '1') and op_break = '0' then
        decode.pc_plus4_d3 <= fetch.pc_plus4_d2;
        decode.pc_plus8_d3 <= std_logic_vector(unsigned(fetch.pc_plus4_d2)+1);  --increment 4
      end if;

      --   if reset='1'or (branch_execute ='1' and stall='0' )then
      if reset = '1'or (branch_execute = '1' and stall_d1 = '0' and
                        (exception_pipeline > "010" or exception_pipeline = "000"))then
        decode.IR_d1               <= (others => '0');
        control_out.restore        <= '0';
        control_out.branch_func    <= (others => '0');
        control_out.arith_func     <= (others => '0');
        control_out.shift_func     <= (others => '0');
        control_out.logic_func     <= (others => '0');
        control_out.mult_func      <= (others => '0');
        control_out.data_in        <= (others => '0');
        control_out.dataflow       <= (others => '0');
        control_out.mem_source_out <= (others => '0');
        
      end if;
    end if;
  end process;


  decode.arith_func  <= control_out.arith_func;
  decode.shift_func  <= control_out.shift_func;
  decode.logic_func  <= control_out.logic_func;
  decode.mult_func   <= control_out.mult_func;
  decode.data_in     <= control_out.data_in;
  decode.branch_func <= control_out.branch_func;
  decode.dataflow    <= control_out.dataflow;
  decode.mem_source  <= control_out.mem_source_out;

--32 general registers
  Inst_reg_bank : reg_bank port map(
    cpu_clk     => cpu_clk,
    stall       => stall,
    BREADY      => BREADY,
    rs_index    => fetch.IR(25 downto 21),
    rt_index    => fetch.IR(20 downto 16),
    regrs_out   => decode.reg_rs_data,
    regrt_out   => decode.reg_rt_data,
    execute_reg => execute_reg
    );

end Behavioral;
