---------------------------------------------------------------------
-- TITLE: Instruction execute
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 27.6.2011
-- FILENAME: inst_execute.vhd
-- PROJECT: MAIS CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
-- 
-- DESCRIPTION:
-- Implements pipeline stage execute
---------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


use work.mlite_pack.all;
use work.alu_pack.all;
use work.branch_pack.all;
use work.SPB.all;

entity inst_execute is
  generic(multiplier : string := "generic");
  port (

    cpu_clk    : in  std_logic;
    mult_ready : out std_logic := '0';
    reset      : in  std_logic;
    stall      : in  std_logic;
    BREADY     : in  std_logic;
    decode     : in  decode_rec;
    loadto_reg : in  write_reg_rec;
    --CP0

    CP0_out    : in  CP0_rec;
    CPint      : in  CPint_rec;
    CP0_write  : out write_reg_rec;

--fetch    
    branch      : out branch_rec;    --! control the program counter

    execute_reg : out write_reg_rec;
--load_access

    IR_load     : out std_logic_vector (2 downto 0);
    load_index  : out std_logic_vector (4 downto 0);

--BBUS
    bus_wait    : in     std_logic;
    BADDR       : out    std_logic_vector (31 downto 0);
    BSEL        : out    std_logic_vector (sel_size-1 downto 0);
    BRD         : out    std_logic;
    BWR         : out    std_logic;
    BACTIVE     : out    std_logic;
    BMASK       : out    std_logic_vector (3 downto 0);
    BWDATA      : out    std_logic_vector (31 downto 0)
   
    );
end inst_execute;

architecture Behavioral of inst_execute is
  
  
  signal mult_akku : std_logic_vector(63 downto 0) := (others => '0');

  signal alu_out        : std_logic_vector(31 downto 0);
  signal path_a         : std_logic_vector(31 downto 0);
  signal path_b         : std_logic_vector(31 downto 0);
  signal compare        : compare_type;
  signal rt_data        : std_logic_vector(31 downto 0);
  signal wr_reg_index   : std_logic_vector(4 downto 0) := (others => '0');
  signal wr_reg_data    : std_logic_vector(31 downto 0);
  signal d1_load_index  : std_logic_vector(4 downto 0);
  signal branch_buf     : branch_rec;
  signal wr_reg         : std_logic;
  signal pc_plus4_d4    : std_logic_vector (31 downto 2);


  signal addr_low       : std_logic_vector(1 downto 0);
  signal bus_active     : std_logic;



  component multiplier_HW
    port (
      cpu_clk    : in  std_logic;
      reset      : in  std_logic;
      path_a     : in  std_logic_vector(31 downto 0);
      path_b     : in  std_logic_vector(31 downto 0);
      decode     : in  decode_rec;
      mult_ready : out std_logic                     := '0';
      mult_akku  : out std_logic_vector(63 downto 0) := (others => '0')
      );
  end component;
  
begin

   BACTIVE<=bus_active;
  
  mult : if (multiplier /= "none") generate
    
    inst_mult : multiplier_HW port map(
      cpu_clk    => cpu_clk,
      reset      => reset,
      path_a     => path_a,
      path_b     => path_b,
      decode     => decode,
      mult_ready => mult_ready,
      mult_akku  => mult_akku
      );

  end generate mult;


  alu_out <= alu_f(path_a, path_b, rt_data, decode.arith_func,
                   decode.shift_func, decode.logic_func, decode.dataflow);


  process (cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if stall = '0' then
        CP0_write.index <= decode.IR_d1(15 downto 11);
        CP0_write.data  <= rt_data;

        if decode.IR_d1(31 downto 21) = "01000000100" then  --op mtcp0
          CP0_write.wr <= '1';
        else
          CP0_write.wr <= '0';
        end if;
      end if;
    end if;
  end process;




  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if stall = '0' then
        branch_buf <= branch_f(decode.branch_func, decode.PC_branch, decode.IR_d1, path_a, decode.pc_plus4_d3, compare, CPint);
      end if;
    end if;
  end process;


------- wichtig hier
-- path_a= register(rs)
  path_a <= mult_akku(31 downto 0) when decode.mult_func(0) = '1' else
            mult_akku(63 downto 32) when decode.mult_func(2) = '1'            else
            X"00000000"             when decode.IR_d1(25 downto 21) = "00000" else
            wr_reg_data             when decode.IR_d1(25 downto 21) = wr_reg_index and wr_reg = '1'
            else
            decode.reg_rs_data;
--multiplexer



-- premupltiplexer path_b
  rt_data <=
    X"00000000"     when decode.IR_d1(20 downto 16) = "00000" else
    wr_reg_data     when decode.IR_d1(20 downto 16) = wr_reg_index and wr_reg = '1'
    else
    decode.reg_rt_data;

  path_b <=
    decode.imm32    when decode.data_in(1) = '1' else
    X"00000000"     when decode.data_in(2) = '1' else  --branch
    rt_data;



--compare
  process (path_a, path_b)
  begin

    if signed(path_a) = signed(path_b) then
      compare(1) <= '1';
    else
      compare(1) <= '0';
    end if;

    if signed(path_a) > signed(path_b) then
      compare(2) <= '1';
    else
      compare(2) <= '0';
    end if;
     
  end process;

-- register index
  process(cpu_clk)
    variable index : std_logic_vector (4 downto 0);
  begin
    if rising_edge(cpu_clk) then
      index := "00000";

      if decode.data_in(1) = '1' then
        index := decode.IR_d1(20 downto 16);  -->rt
      end if;

      if decode.data_in(0) = '1' then
        index := decode.IR_d1(15 downto 11);        -->rd
      end if;
      if decode.data_in(6 downto 3) /= "0000" then  --CP0-3
        index := decode.IR_d1(20 downto 16);        -->rt
      end if;
      
      if decode.branch_func(5) = '1' then
        index := "11111";               -- link in R[31]
      end if;
      if loadto_reg.wr = '1' then       --read from RAM
        index := loadto_reg.index;
      end if;
      wr_reg_index <= index;
    end if;
  end process;


  process(cpu_clk)
    variable data, data1 : std_logic_vector(31 downto 0);
  begin
    if rising_edge(cpu_clk) then

      data := alu_f(path_a, path_b, rt_data, decode.arith_func,
                    decode.shift_func, decode.logic_func, decode.dataflow);

      data1 := data;
      if decode.data_in(3) = '1' then   -- move from CP0
        data1 := CP0_out.reg;
      end if;
      if decode.branch_func(8) = '1' or decode.branch_func(5) = '1' then
        data1 := decode.pc_plus8_d3 & "00";
      end if;
      if loadto_reg.wr = '1' then
        data1 := loadto_reg.data;
      end if;
      if decode.mult_func(0) = '1' then
        data1 := mult_akku(31 downto 0);
      end if;
      if decode.mult_func(2) = '1' then
        data1 := mult_akku(63 downto 32);
      end if;  

      wr_reg_data <= data1;
      
    end if;
  end process;


  process (cpu_clk)
    variable wr : std_logic;
  begin
    if rising_edge(cpu_clk) then
      wr := '0';
      if loadto_reg.wr = '1' then       --read from RAM
        wr := '1';
      end if;
      if stall = '0' and bus_wait = '0' then
        if decode.dataflow(3) = '1' then
          wr := '1';
        end if;
        if decode.data_in(3) = '1' then  -- move from CP0
          wr := '1';
        end if;
        if decode.branch_func(5) = '1' then
          if decode.branch_func(4 downto 3) = "11" and path_a(31) = '0' then  --BGEZAL
            wr := '1';
          end if;
          if decode.branch_func(2) = '1' and path_a(31) = '1' then  --BLTZAL
            wr := '1';
          end if;
        end if;
        if decode.branch_func(6 downto 5) = "11" then               --jal
          wr := '1';
        end if;
        if decode.branch_func(8 downto 7) = "11" then               --jalr
          wr := '1';
        end if;

        if decode.mult_func(0) = '1' or decode.mult_func(2) = '1'then
          wr := '1';
        end if;
      end if;
      wr_reg <= wr;
    end if;
  end process;



  execute_reg.index <= wr_reg_index;
  execute_reg.data  <= wr_reg_data;
  execute_reg.wr    <= wr_reg;
  branch            <= branch_buf;



-----------------------------------------------------------------
-- Bus 
--    
-- the register index for load



  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
     -- if not (BREADY = '0' and BACTIVE = '1') then
     BUS_ACTIVE<='0';
     BWR<='0';
     BRD<='0';     
     if BREADY='1' then
        if stall = '0'and decode.mem_source(7) = '1' then
          BUS_ACTIVE <= '1';
          BSEL   <= Bsel_f (path_a(31 downto 28));
          BADDR  <= alu_out;
          if decode.mem_source(6) = '1' then
            load_index <= decode.IR_d1(20 downto 16);
            BWR  <='0';
            BRD  <='1';
          else
            BWR<='1';
            BRD<='0';
          end if;
        end if;
      end if;

      if reset = '1' then
        BSEL   <= (others => '0');
        BADDR  <= (others => '0');
	    BUS_ACTIVE <= '0';
        BWR <='0';
        BRD <='0';
      end if;
    end if;
  end process;

  
  addr_low <= std_logic_vector(unsigned(path_a(1 downto 0))
                               +unsigned(decode.IR_d1(1 downto 0)));


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if decode.mem_source(7) = '1' and stall = '0' then
        case addr_low(1 downto 0) is    --8bit
          when "00"   => BMASK <= "1000";
          when "01"   => BMASK <= "0100";
          when "10"   => BMASK <= "0010";
          when "11"   => BMASK <= "0001";
          when others => null;
        end case;

        if decode.mem_source(1) = '1' then  --16bit
          if addr_low(1) = '0' then
            BMASK <= "1100";
          else
            BMASK <= "0011";
          end if;
        end if;

        if decode.mem_source(2) = '1' then  --  32 bit
          BMASK <= "1111";
        end if;



        if decode.mem_source(5) = '1' then  --word right
          case addr_low(1 downto 0) is
            when "00"   => BMASK <= "1000";
            when "01"   => BMASK <= "1100";
            when "10"   => BMASK <= "1110";
            when "11"   => BMASK <= "1111";
            when others => BMASK <= "0000";
          end case;
        end if;
        if decode.mem_source(4) = '1' then  --word left
          case addr_low(1 downto 0) is
            when "00"   => BMASK <= "1111";
            when "01"   => BMASK <= "0111";
            when "10"   => BMASK <= "0011";
            when "11"   => BMASK <= "0001";
            when others => BMASK <= "0000";
          end case;
        end if;
      end if;
    end if;
  end process;



--store
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if not (BREADY = '0' and BUS_ACTIVE = '1') then
        if decode.mem_source(7) = '1' and stall = '0' then
          BWDATA <= rt_data;                 --32bit 
          if decode.mem_source(0) = '1' then  -- 8bit
            BWDATA <= rt_data (7 downto 0) & rt_data (7 downto 0)&
                       rt_data (7 downto 0) & rt_data (7 downto 0);
          end if;
          if decode.mem_source(1) = '1' then  --16bit
            BWDATA <= rt_data (15 downto 0) & rt_data (15 downto 0);
          end if;

          if decode.mem_source(5) = '1' and decode.mem_source(6) = '0' then  --SWR
            case addr_low(1 downto 0) is
              when "00" => BWDATA <= rt_data(7 downto 0) & X"000000";
              when "01" => BWDATA <= rt_data(15 downto 0)& X"0000";
              when "10" => BWDATA <= rt_data(23 downto 0)& X"00";

              when others => BWDATA <= rt_data;  --32bit;        
            end case;
          end if;
          if decode.mem_source(4) = '1' and decode.mem_source(6) = '0' then  --SWL
            case addr_low(1 downto 0) is
              when "01"   => BWDATA <= X"00" & rt_data(31 downto 8);
              when "10"   => BWDATA <= X"0000" & rt_data(31 downto 16);
              when "11"   => BWDATA <= X"000000" & rt_data(31 downto 24);
              when others => BWDATA <= rt_data;  --32bit  
            end case;
          end if;
        end if;
      end if;
    end if;
  end process;


  process(cpu_clk)
    variable store_b : std_logic_vector(2 downto 0);
  begin
    if rising_edge(cpu_clk) then
      ir_load <= decode.IR_d1(28 downto 26);
    end if;    
  end process;


end Behavioral;

