---------------------------------------------------------------------
-- TITLE: Instruction fetch
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 27.6.2011
-- FILENAME: inst_fetch.vhd
-- PROJECT: MAIS CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
-- 
-- DESCRIPTION:
-- Implements the Program Counter logic and fetch the instruction
---------------------------------------------------------------------
--! @file inst_fetch.vhd
--! @brief implementation of program counter logic and fetch the instruction
--! @author Rene Doss
--! @version 0.9
---------------------------------------------------------------------
--! \image html html/inst_fetch.png


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mlite_pack.all;


entity inst_fetch is
  port(cpu_clk            : in std_logic;  --! cpu clock
       reset              : in std_logic;  --! reset  
       syscall            : in std_logic;
       break7             : in std_logic;
       exception          : in std_logic;
       exception_pipeline : in unsigned (2 downto 0);
       --bus activity           
       bus_wait           : in std_logic;
       BACTIVE            : in std_logic;
       BREADY             : in std_logic;

       --pc register       
       pc_reg             : in std_logic_vector(31 downto 0);
       pc_plus4_d1        : in std_logic_vector(31 downto 0);
       
       --instruction memory
       imem_data          : in std_logic_vector(31 downto 0);
       imem_valid         : in std_logic;
       ---------------------------- 
       mult_ready         : in     std_logic;
       branch             : in     branch_rec;  --! control IR 
       stall              : in     std_logic;
       stall_mult         : buffer std_logic := '0';

       op_break           : in std_logic;  --! hold pc and put nop into instruction register      
       debug_fetch_hold   : in std_logic;
       fetch : out fetch_rec); 
end;


architecture Behavioral of inst_fetch is

  signal reset_a, reset_b  : std_logic                      := '0';  --! reset catcher
  signal ir , ir_stall     : std_logic_vector (31 downto 0) := (others => '0');  --! instruction re
  signal stall_d1          : std_logic;
  signal branch_execute_d1 : std_logic;
  signal stall_mult_d1     : std_logic                      := '0';
  signal syscall_d1        : std_logic;
  signal break_exc         : std_logic;

  
begin


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      reset_a <= reset;
      reset_b <= reset_a;
      if exception_pipeline(2 downto 0) = "001" then
        branch_execute_d1 <= '0';
      else
        branch_execute_d1 <= branch.branch_execute;
      end if;
    end if;
  end process;

  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      stall_mult_d1 <= stall_mult;
      --op  mult/multu/div/divu 
      if IR(31 downto 26) = "000000" and IR(5 downto 3) = "011" then
        stall_mult <= '1';
      end if;
      if mult_ready = '1' then
        stall_mult <= '0';
      end if;
      if reset = '1' then
        stall_mult <= '0';
      end if;
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      syscall_d1 <= syscall;
    end if;
  end process;

  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then     
      fetch.IR_new <= '0';
      if reset_b = '1' or ((stall_d1 = '0' and
                            (branch.branch_execute = '1' or branch_execute_d1 = '1')) and (exception_pipeline="000"))
        or syscall = '1' or syscall_d1 = '1' then
        IR <= X"00000000";  --opcode 0x00000000 is instruction NOP!!
      else
        if imem_valid = '1' then
          IR           <= imem_data;
          fetch.IR_new <= '1';
        else
          IR           <= X"00000000";  --opcode 0x00000000 is instruction NOP!!
          fetch.IR_new <= '0';
        end if;
        if stall = '1' or stall_mult = '1' or (BACTIVE = '1' and BREADY = '0')then
          IR           <= IR;
          fetch.IR_new <= '0';
        end if;
          

        if (stall = '0' and stall_d1 = '1') or(stall_mult = '0' and stall_mult_d1 = '1') or
          (bus_wait = '1' and BREADY = '1') then
          IR           <= IR_stall;
          fetch.IR_new <= '1';
        end if;


        if IR(31 downto 26) = "000000" and IR(5 downto 1) = "00110" and exception = '0' then
          IR           <= IR;
          fetch.IR_new <= '1';
        end if;

        if (op_break = '1' and exception = '1') or break_exc = '1' then
          IR           <= X"00000000";
          fetch.IR_new <= '1';
        end if;

        if debug_fetch_hold = '1'  then
          IR           <= X"00000000";  --opcode 0x00000000 is instruction NOP!!  
          fetch.IR_new <= '1';
        end if;
      end if;
      if break7='1' then             --rescoure op ater break for multiplier
          IR           <= IR_stall;
          fetch.IR_new <= '0';
        end if;      
    end if;
  end process;


  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if op_break = '1' and exception = '1' then
        break_exc <= '1';
      else
        break_exc <= '0';
      end if;
    end if;
  end process;



  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      stall_d1 <= stall;
      if (stall = '0' or (stall = '1' and stall_d1 = '0')) and stall_mult_d1 = '0'and BUS_wait = '0' then
        
        if imem_valid = '1' then
          IR_stall <= imem_data;
        else
          IR_stall <= X"00000000";  --opcode 0x00000000 is instruction NOP!!
        end if;
      end if;
      if branch_execute_d1 = '1' then
        IR_stall <= X"00000000";    --opcode 0x00000000 is instruction NOP!!
      end if;
    end if;
  end process;


-- This is a delay because blockram is sync reading
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if stall = '1' or op_break = '1' or stall_mult = '1' then
      else
        fetch.pc_d1       <= pc_reg (31 downto 2);
        fetch.pc_plus4_d2 <= PC_plus4_d1(31 downto 2);
      end if;
    end if;
  end process;

  fetch.IR <= IR;

end;  --Behavioral
