---------------------------------------------------------------------
-- TITLE: alu
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 2/8/01
-- FILENAME: alu.vhd
-- PROJECT: MIPS CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--
-- DESCRIPTION:
--    Implement the operations shift,add,or,and.......
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mlite_pack.all;


package alu_pack is
  function alu_f (
    a          : in std_logic_vector(31 downto 0);
    b          : in std_logic_vector(31 downto 0);
    rt_data    : in std_logic_vector(31 downto 0);
    arith_func : in arith_function_type;
    shift_func : in shift_function_type;
    logic_func : in logic_function_type;
    dataflow   : in dataflow_type)
    return std_logic_vector;

  function arith_f
    (
      a          : in std_logic_vector(31 downto 0);
      b          : in std_logic_vector(31 downto 0);
      arith_func : in arith_function_type)
    return std_logic_vector;
  
  function logic_f
    (a          : in std_logic_vector(31 downto 0);
     b          : in std_logic_vector(31 downto 0);
     logic_func : in logic_function_type)
    return std_logic_vector;
  
  
  function shift_f
    (a          : in std_logic_vector(31 downto 0);
     b          : in std_logic_vector(31 downto 0);
     shift_func : in shift_function_type;
     rt_data    : in std_logic_vector(31 downto 0)) 
    return std_logic_vector;

end;

package body alu_pack is

  
  function alu_f (
    a          : in std_logic_vector(31 downto 0);
    b          : in std_logic_vector(31 downto 0);
    rt_data    : in std_logic_vector(31 downto 0);
    arith_func : in arith_function_type;
    shift_func : in shift_function_type;
    logic_func : in logic_function_type;
    dataflow   : in dataflow_type)
    return std_logic_vector
  is
    variable arith   : std_logic_vector(31 downto 0);
    variable shift   : std_logic_vector(31 downto 0);
    variable logic   : std_logic_vector(31 downto 0);
    variable alu_out : std_logic_vector(31 downto 0);
    variable slt_v   : std_logic;
    variable sltu_v  : std_logic;
    
  begin
    
    arith := arith_f(a, b, arith_func);
    shift := shift_f(a, b, shift_func, rt_data);
    logic := logic_f(a, b, logic_func);



    if signed(a) < signed(b) then
      slt_v := '1';
    else
      slt_v := '0';
    end if;



    if unsigned(a) < unsigned(b) then
      sltu_v := '1';
    else
      sltu_v := '0';
    end if;

    
    alu_out := arith;
    
    if dataflow(1) = '1' then
      alu_out := shift;
    end if;

    if dataflow(2) = '1' then
      alu_out := logic;
    end if;
    if dataflow(4) = '1' then
      alu_out(31 downto 1) := (others => '0');
      alu_out(0)           := slt_v;
    end if;
    if dataflow(5) = '1' then
      alu_out(31 downto 1) := (others => '0');
      alu_out(0)           := sltu_v;
    end if;



    return alu_out;
  end;




  function arith_f
    (
      a          : in std_logic_vector(31 downto 0);
      b          : in std_logic_vector(31 downto 0);
      arith_func : in arith_function_type)
    return std_logic_vector
  is
    variable arith : std_logic_vector(31 downto 0);
  begin
    if arith_func(0) = '1' then
      arith := std_logic_vector(unsigned(a)+unsigned(b));
    else
      arith := std_logic_vector(unsigned(a)-unsigned(b));
    end if;
    return arith;
  end arith_f;

  function logic_f
    (a          : in std_logic_vector(31 downto 0);
     b          : in std_logic_vector(31 downto 0);
     logic_func : in logic_function_type)
    return std_logic_vector
  is
    variable logic : std_logic_vector(31 downto 0);
  begin
    if logic_func(0) = '1' then
      logic := a or b;
    end if;
    if logic_func(1) = '1' then
      logic := a and b;
    end if;
    if logic_func(2) = '1' then
      logic := a xor b;
    end if;
    if logic_func(3) = '1' then
      logic := a nor b;
    end if;
    if logic_func(4) = '1' then         --load upper imediate
      logic := b(15 downto 0) & x"0000";
    end if;
    return logic;
  end logic_f;


  function shift_f
    (a          : in std_logic_vector(31 downto 0);
     b          : in std_logic_vector(31 downto 0);
     shift_func : in shift_function_type;
     rt_data    : in std_logic_vector(31 downto 0)) 
    return std_logic_vector
    
  is

    --barrel shifter unit
    variable shift1L, shift2L, shift4L, shift8L, shift16 :
      std_logic_vector(31 downto 0);
    variable shift1R, shift2R, shift4R, shift8R :
      std_logic_vector(31 downto 0);
    variable fills       : std_logic_vector(31 downto 16);
    variable shiftnumber : std_logic_vector(4 downto 0);
  begin

    if (shift_func(1) = '1' and rt_data(31) = '1') then
      fills := "1111111111111111";
    else
      fills := "0000000000000000";
    end if;

    if shift_func(2) = '0' then
      shiftnumber := b(10 downto 6);
    else
      shiftnumber := a(4 downto 0);
    end if;
    --SLL  rd<_rt << sa
    if shift_func (0) = '0' then        -- shift left
      if shiftnumber(0) = '1' then
        shift1L := rt_data(30 downto 0) & '0';
      else
        shift1L := rt_data;
      end if;
      if shiftnumber(1) = '1' then
        shift2L := shift1L(29 downto 0) & "00";
      else
        shift2L := shift1L;
      end if;
      if shiftnumber(2) = '1' then
        shift4L := shift2L(27 downto 0) & "0000";
      else
        shift4L := shift2L;
      end if;
      if shiftnumber(3) = '1' then
        shift8L := shift4L(23 downto 0) & "00000000";
      else
        shift8L := shift4L;
      end if;
      if shiftnumber(4) = '1' then
        shift16 := shift8L(15 downto 0) & X"0000";
      else
        shift16 := shift8L;
      end if;
      
    else                                --shift right
      if shiftnumber(0) = '1' then
        shift1R := fills(31) & rt_data(31 downto 1);
      else
        shift1R := rt_data;
      end if;
      if shiftnumber(1) = '1' then
        shift2R := fills(31 downto 30) & shift1R(31 downto 2);
      else
        shift2R := shift1R;
      end if;
      if shiftnumber(2) = '1' then
        shift4R := fills(31 downto 28) & shift2R(31 downto 4);
      else
        shift4R := shift2R;
      end if;
      if shiftnumber(3) = '1' then
        shift8R := fills(31 downto 24) & shift4R(31 downto 8);
      else
        shift8R := shift4R;
      end if;
      if shiftnumber(4) = '1' then
        shift16 := fills(31 downto 16) & shift8R(31 downto 16);
      else
        shift16 := shift8R;
      end if;
    end if;
    return(shift16);
    
  end shift_f;
  
  
end;
