---------------------------------------------------------------------
-- title: package
-- author: Rene Doss
-- company Dossmatik GmbH
-- date created: 27.6.2011
-- filename: pkg_mlite.vhd
-- project: Mais CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--
---------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


package mlite_pack is
  

  subtype  data_in_type is std_logic_vector (6 downto 0);
-----datainput  
--   bit 0  r[rs]  operand r[rt] -> r[rd] (and MFLO;MFHI)
--   bit 1  r[rs]  operand imm  -> r[rt]
--   bit 2  r[rs]  operand zero
--   bit 3  cp0  input   
--   bit 4  cp1  input
--   bit 5  cp2  input
--   bit 6  cp3  input
  constant reg_reg   : data_in_type := "0000001";
  constant reg_imm   : data_in_type := "0000010";
  constant reg_sa    : data_in_type := "0000011";
  constant rs_zero   : data_in_type := "0000100";
  constant cp0_input : data_in_type := "0001000";
  constant cp1_input : data_in_type := "0010000";
  constant cp2_input : data_in_type := "0100000";
  constant cp3_input : data_in_type := "1000000";




  subtype dataflow_type is std_logic_vector (7 downto 0);
-----dataflow  aluoutput
--   bit 0  arith
--   bit 1  shift
--   bit 2  logic  
--------dataout
--   bit 3  alu -> register
--   bit 4  set on less/ signed compare  ->register 
--   bit 5  set on less/ unsigned compare->register
--   bit 6  multiply akku
--   bit 7  addresspointer  (alu->addess dram)

  subtype arith_function_type is std_logic_vector (3 downto 0);
--   bit 0  add   (add +addu)
--   bit 1  sub   (sub+subu)
--   bit 2  count leading ones (rs)
--   bit 3  count leading zeros(rs)

  subtype shift_function_type is std_logic_vector (2 downto 0);
--   bit 0  direction 
--       ='1'    shift right
--       ='0'    shift left
--   shift type 
--   bit 1  ='0' logic  ='1' arith
--   bit 2  ='1' shift r[rs]  ='0'  shift op(10 downto 6)   

  subtype logic_function_type is std_logic_vector (4 downto 0);
--   bit 0  or
--   bit 1  and
--   bit 2  xor
--   bit 3  nor
--   bit 4  lui load upper immediate
--   bit 13 slti
--   bit 14 sltiu

  subtype compare_type is std_logic_vector (2 downto 1);
--   path_a  rs   
--   path_b  rt     
--   bit 0 r[rs] path_a < path_b
--   bit 1 r[rs] path_a = path_b
--   bit 2 r[rs] path_a > path_b
--   bit 3 r[rs] unsigned(path_a)<unsigned(path_b)

  subtype mult_function_type is std_logic_vector (7 downto 0);

  constant mult_read_low  : mult_function_type := "00000001";
  constant mult_write_low : mult_function_type := "00000010";
  constant mult_read_hi   : mult_function_type := "00000100";
  constant mult_write_hi  : mult_function_type := "00001000";
  constant mult_mult      : mult_function_type := "01010000";
  constant mult_multu     : mult_function_type := "01100000";
  constant mult_div       : mult_function_type := "10010000";
  constant mult_divu      : mult_function_type := "10100000";

  subtype branch_function_type is std_logic_vector (10 downto 0);
--   bit 10 eret (return from interrupt service routine)
--   bit 9  likely  clear branch delay slot  
--   bit 8  link r[rd]<-[pc+8]
--   bit 7  jump  rs
--   bit 6  jump  op(25 downto 0) in the 256mb aligned region
--   bit 5  is link r[31]<-[pc+8]
--   bit 4  [rs]>0
--   bit 3  [rs]=0 
--   bit 2  [rs]<0
--   bit 1  compare rs and rt
--   bit 0  1= equal  0=not equal  

  constant branch_no : branch_function_type := "00000000000";

  constant eret : branch_function_type := "10000000000";

-- table a-12 jump instructions                            9876543210
  constant jump                : branch_function_type := "00001000000";
-- jump uncondition  j
  constant jump_al             : branch_function_type := "00001100000";
-- jump uncondition and link  jal
-- table a13 jump absolute
  constant jump_rs             : branch_function_type := "00010000000";
-- jump register jr
  constant jump_al_rs          : branch_function_type := "00110000000";
-- jump uncondition and link [rs] jalr   
-- table a-14 two register comparing instruction 
  constant branch_eq           : branch_function_type := "00000000011";
-- r[rs]=[rt] beq
  constant branch_ne           : branch_function_type := "00000000010";
-- r[rs]<>[rt] bne
  constant branch_lez          : branch_function_type := "00000001100";
-- r[rs] =<0 blez
  constant branch_gtz          : branch_function_type := "00000010000";
-- r[rs]>  0 bgtz
  constant branch_eq_likely    : branch_function_type := "01000000011";
-- r[rs]=[rt] beql likely  
  constant branch_ne_likely    : branch_function_type := "01000000010";
-- r[rs]<>[rt] bnel likely 
  constant branch_lez_likely   : branch_function_type := "01000001100";
-- r[rs] =<0 blezl likely   
  constant branch_gtz_likely   : branch_function_type := "01000010000";
-- r[rs]>  0 bgtzl likely  
-- table a-15                                              9876543210
  constant branch_ltz          : branch_function_type := "00000000100";
-- r[rs]  <0  bltz
  constant branch_gez          : branch_function_type := "00000011000";
-- r[rs]>= 0  bgez
  constant branch_ltzal        : branch_function_type := "00000100100";
-- r[rs]  <0 and link bltzal 
  constant branch_gezal        : branch_function_type := "00000111000";
-- r[rs]>= 0 and link bgezal
  constant branch_ltz_likely   : branch_function_type := "01000000100";
-- r[rs]  <0 likely  bltzl
  constant branch_gez_likely   : branch_function_type := "01000011000";
-- r[rs]>= 0 likely  bgezl
  constant branch_ltzal_likely : branch_function_type := "01000011000";
-- r[rs]  <0 and link likely bltzall
  constant branch_gezal_likely : branch_function_type := "01000111000";
-- r[rs]>= 0 and link likely bgezall



-- mode(32=1,16=2,8=3), signed, write
  subtype mem_source_type is std_logic_vector (7 downto 0);
--  bit 7  1 data-ram access
--  bit 6  1=load  0=store
--  bit 5  word right
--  bit 4  word left
--  bit 3  1 signed     0 unsigned
--  bit 2  32 bit
--  bit 1  16 bit
--  bit 0   8 bit

  type CP0_rec is
  record
    reg : std_logic_vector (31 downto 0);
  end record;

  type cpint_rec is
  record
    epc : std_logic_vector (31 downto 2);
  end record;

  type control_rec is
  record
    syscall        : std_logic;
    restore        : std_logic;
    arith_func     : arith_function_type;
    shift_func     : shift_function_type;
    logic_func     : logic_function_type;
    mult_func      : mult_function_type;

    data_in        : data_in_type;
    branch_func    : branch_function_type;
    dataflow       : dataflow_type;
    mem_source_out : mem_source_type;
  end record;

  type write_reg_rec is
  record
    wr    : std_logic;
    index : std_logic_vector (4 downto 0);
    data  : std_logic_vector (31 downto 0);
  end record;

  type fetch_rec is
  record
    ir          : std_logic_vector (31 downto 0);
    ir_new      : std_logic;
--instruction reg
    pc_d1       : std_logic_vector (31 downto 2);
    pc_plus4_d2 : std_logic_vector (31 downto 2);
--next_seq_pc           
  end record;
  type branch_rec is
  record
    branch_execute : std_logic;
    branch_address : std_logic_vector (31 downto 2);
  end record;


  type decode_rec is
  record
    pc_plus4_d3 : std_logic_vector (31 downto 2);
    pc_plus8_d3 : std_logic_vector (31 downto 2);
    pc_branch   : std_logic_vector (31 downto 2);
    reg_rs_data : std_logic_vector (31 downto 0);
    reg_rt_data : std_logic_vector (31 downto 0);
    imm32       : std_logic_vector (31 downto 0);
    ir_d1       : std_logic_vector (31 downto 0);

    syscall     : std_logic;
    restore     : std_logic;
    arith_func  : arith_function_type;
    shift_func  : shift_function_type;
    logic_func  : logic_function_type;
    mult_func   : mult_function_type;
    data_in     : data_in_type;
    branch_func : branch_function_type;
    dataflow    : dataflow_type;
    mem_source  : mem_source_type;
  end record;

  type t_gpr is
  record
    gpr1  : std_logic_vector (31 downto 0);
    gpr2  : std_logic_vector (31 downto 0);
    gpr3  : std_logic_vector (31 downto 0);
    gpr4  : std_logic_vector (31 downto 0);
    gpr5  : std_logic_vector (31 downto 0);
    gpr6  : std_logic_vector (31 downto 0);
    gpr7  : std_logic_vector (31 downto 0);
    gpr8  : std_logic_vector (31 downto 0);
    gpr9  : std_logic_vector (31 downto 0);
    gpr10 : std_logic_vector (31 downto 0);
    gpr11 : std_logic_vector (31 downto 0);
    gpr12 : std_logic_vector (31 downto 0);
    gpr13 : std_logic_vector (31 downto 0);
    gpr14 : std_logic_vector (31 downto 0);
    gpr15 : std_logic_vector (31 downto 0);
    gpr16 : std_logic_vector (31 downto 0);
    gpr17 : std_logic_vector (31 downto 0);
    gpr18 : std_logic_vector (31 downto 0);
    gpr19 : std_logic_vector (31 downto 0);
    gpr20 : std_logic_vector (31 downto 0);
    gpr21 : std_logic_vector (31 downto 0);
    gpr22 : std_logic_vector (31 downto 0);
    gpr23 : std_logic_vector (31 downto 0);
    gpr24 : std_logic_vector (31 downto 0);
    gpr25 : std_logic_vector (31 downto 0);
    gpr26 : std_logic_vector (31 downto 0);
    gpr27 : std_logic_vector (31 downto 0);
    gpr28 : std_logic_vector (31 downto 0);
    gpr29 : std_logic_vector (31 downto 0);
    gpr30 : std_logic_vector (31 downto 0);
    gpr31 : std_logic_vector (31 downto 0);
  end record;

end;


package body mlite_pack is

 
  
end;

