---------------------------------------------------------------------
-- TITLE: Instruction fetch
-- AUTHOR: Rene Doss
-- Company Dossmatik GmbH
-- DATE CREATED: 27.6.2011
-- FILENAME: reg_bank.vhd
-- PROJECT: MAIS CPU core
--
-- COPYRIGHT: Creative Commons CC BY-NC 3.0 with exception 
-- commercial applicants have to pay a licence fee
--
-- DESCRIPTION:
--    Implements a register bank with 32 registers that are 32-bits wide.
--    There are two read-ports and one write port.
---------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mlite_pack.all;


entity reg_bank is
  port(cpu_clk     : in  std_logic;
       rs_index    : in  std_logic_vector(4 downto 0);
       rt_index    : in  std_logic_vector(4 downto 0);
       stall       : in  std_logic;
       BREADY      : in  std_logic;
       regrs_out   : out std_logic_vector(31 downto 0);
       regrt_out   : out std_logic_vector(31 downto 0);
       execute_reg : in  write_reg_rec);
end;


architecture ram_block of reg_bank is
  signal intr_enable_reg : std_logic;

  signal gpr : t_gpr;


begin

--read Register[rs]
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if stall = '0' and BREADY = '1' then
--Zero register is shadowed in the multiplexer in path_a and path_b
        if rs_index = execute_reg.index and execute_reg.wr = '1' then
          regrs_out <= execute_reg.data;
        else
          
          case rs_index is
            when "00000" => regrs_out <= (others => '0');
            when "00001" => regrs_out <= gpr.gpr1;
            when "00010" => regrs_out <= gpr.gpr2;
            when "00011" => regrs_out <= gpr.gpr3;
            when "00100" => regrs_out <= gpr.gpr4;
            when "00101" => regrs_out <= gpr.gpr5;
            when "00110" => regrs_out <= gpr.gpr6;
            when "00111" => regrs_out <= gpr.gpr7;
            when "01000" => regrs_out <= gpr.gpr8;
            when "01001" => regrs_out <= gpr.gpr9;
            when "01010" => regrs_out <= gpr.gpr10;
            when "01011" => regrs_out <= gpr.gpr11;
            when "01100" => regrs_out <= gpr.gpr12;
            when "01101" => regrs_out <= gpr.gpr13;
            when "01110" => regrs_out <= gpr.gpr14;
            when "01111" => regrs_out <= gpr.gpr15;
            when "10000" => regrs_out <= gpr.gpr16;
            when "10001" => regrs_out <= gpr.gpr17;
            when "10010" => regrs_out <= gpr.gpr18;
            when "10011" => regrs_out <= gpr.gpr19;
            when "10100" => regrs_out <= gpr.gpr20;
            when "10101" => regrs_out <= gpr.gpr21;
            when "10110" => regrs_out <= gpr.gpr22;
            when "10111" => regrs_out <= gpr.gpr23;
            when "11000" => regrs_out <= gpr.gpr24;
            when "11001" => regrs_out <= gpr.gpr25;
            when "11010" => regrs_out <= gpr.gpr26;
            when "11011" => regrs_out <= gpr.gpr27;
            when "11100" => regrs_out <= gpr.gpr28;
            when "11101" => regrs_out <= gpr.gpr29;
            when "11110" => regrs_out <= gpr.gpr30;
            when "11111" => regrs_out <= gpr.gpr31;
            when others  => null;
          end case;
        end if;
      end if;
    end if;
    
  end process;



--read Register[rt]
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if stall = '0' and BREADY = '1' then
--Zero register is shadowed in the multiplexer in path_a and path_b
        if rt_index = execute_reg.index and execute_reg.wr = '1' then
          regrt_out <= execute_reg.data;
        else
          case rt_index is
            when "00000" => regrt_out <= (others => '0');
            when "00001" => regrt_out <= gpr.gpr1;
            when "00010" => regrt_out <= gpr.gpr2;
            when "00011" => regrt_out <= gpr.gpr3;
            when "00100" => regrt_out <= gpr.gpr4;
            when "00101" => regrt_out <= gpr.gpr5;
            when "00110" => regrt_out <= gpr.gpr6;
            when "00111" => regrt_out <= gpr.gpr7;
            when "01000" => regrt_out <= gpr.gpr8;
            when "01001" => regrt_out <= gpr.gpr9;
            when "01010" => regrt_out <= gpr.gpr10;
            when "01011" => regrt_out <= gpr.gpr11;
            when "01100" => regrt_out <= gpr.gpr12;
            when "01101" => regrt_out <= gpr.gpr13;
            when "01110" => regrt_out <= gpr.gpr14;
            when "01111" => regrt_out <= gpr.gpr15;
            when "10000" => regrt_out <= gpr.gpr16;
            when "10001" => regrt_out <= gpr.gpr17;
            when "10010" => regrt_out <= gpr.gpr18;
            when "10011" => regrt_out <= gpr.gpr19;
            when "10100" => regrt_out <= gpr.gpr20;
            when "10101" => regrt_out <= gpr.gpr21;
            when "10110" => regrt_out <= gpr.gpr22;
            when "10111" => regrt_out <= gpr.gpr23;
            when "11000" => regrt_out <= gpr.gpr24;
            when "11001" => regrt_out <= gpr.gpr25;
            when "11010" => regrt_out <= gpr.gpr26;
            when "11011" => regrt_out <= gpr.gpr27;
            when "11100" => regrt_out <= gpr.gpr28;
            when "11101" => regrt_out <= gpr.gpr29;
            when "11110" => regrt_out <= gpr.gpr30;
            when "11111" => regrt_out <= gpr.gpr31;
            when others  => null;
          end case;
        end if;
      end if;
    end if;
  end process;

--write
  process(cpu_clk)
  begin
    if rising_edge(cpu_clk) then
      if execute_reg.wr = '1' then
        case execute_reg.index is
          --       when "00000" => null;
          when "00001" => gpr.gpr1  <= execute_reg.data;
          when "00010" => gpr.gpr2  <= execute_reg.data;
          when "00011" => gpr.gpr3  <= execute_reg.data;
          when "00100" => gpr.gpr4  <= execute_reg.data;
          when "00101" => gpr.gpr5  <= execute_reg.data;
          when "00110" => gpr.gpr6  <= execute_reg.data;
          when "00111" => gpr.gpr7  <= execute_reg.data;
          when "01000" => gpr.gpr8  <= execute_reg.data;
          when "01001" => gpr.gpr9  <= execute_reg.data;
          when "01010" => gpr.gpr10 <= execute_reg.data;
          when "01011" => gpr.gpr11 <= execute_reg.data;
          when "01100" => gpr.gpr12 <= execute_reg.data;
          when "01101" => gpr.gpr13 <= execute_reg.data;
          when "01110" => gpr.gpr14 <= execute_reg.data;
          when "01111" => gpr.gpr15 <= execute_reg.data;
          when "10000" => gpr.gpr16 <= execute_reg.data;
          when "10001" => gpr.gpr17 <= execute_reg.data;
          when "10010" => gpr.gpr18 <= execute_reg.data;
          when "10011" => gpr.gpr19 <= execute_reg.data;
          when "10100" => gpr.gpr20 <= execute_reg.data;
          when "10101" => gpr.gpr21 <= execute_reg.data;
          when "10110" => gpr.gpr22 <= execute_reg.data;
          when "10111" => gpr.gpr23 <= execute_reg.data;
          when "11000" => gpr.gpr24 <= execute_reg.data;
          when "11001" => gpr.gpr25 <= execute_reg.data;
          when "11010" => gpr.gpr26 <= execute_reg.data;
          when "11011" => gpr.gpr27 <= execute_reg.data;
          when "11100" => gpr.gpr28 <= execute_reg.data;
          when "11101" => gpr.gpr29 <= execute_reg.data;
          when "11110" => gpr.gpr30 <= execute_reg.data;
          when "11111" => gpr.gpr31 <= execute_reg.data;
          when others  => null;
        end case;
      end if;
    end if;
  end process;

end;
